/* A generic wrapper for programs that can be run by simply exec()'ing the
 * native version.  In general, programs that fit this description are those
 * that do not transfer data. 
 *
 * This wrapper always skips the first node (rank == 0) because of the way that
 * the run wrapper is implemented.  If the user wants to do this operation on the
 * current node as well, he or she need only provide it in the machine list.
 * */

#include "mpi.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

extern char *progname;
extern int check_args(int argc, char *argv[]);

int main(int argc, char *argv[])
{
    int pid;
    int fds[2];
    int rank, status = 0, global_status;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    argv[0] = progname;
    if (rank == 0 && check_args(argc, argv) < 0){
        fprintf(stderr, "Invalid arguments\n");
        status = -1;
    }

    MPI_Bcast(&status, 1, MPI_INT, 0, MPI_COMM_WORLD);
    if (status < 0){
        MPI_Finalize();
        return 1;
    }

    if (rank == 0)
        status = 0;
    else 
        status = pipe(fds);
    MPI_Allreduce(&status, &global_status, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
    if (global_status < 0){
        if (rank == 0)
            fprintf(stderr, "Unable to open pipe\n");
        MPI_Finalize();
        return 1;
    }

    if (rank == 0)
        pid = 1;
    else
        pid = fork();
    /* At each node, only the (failed) parents should reduce */
    if (pid != 0){
        MPI_Allreduce(&pid, &status, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
        /* Notify the child */
        if (rank != 0)
            write(fds[1], &status, sizeof(int));
        if (status < 0){
            if (rank == 0)
                fprintf(stderr, "Fork failed\n");
            MPI_Finalize();
            return 1;
        }
        wait(&status);
    } else {
        /* Check that everyone is ok */
        read(fds[0], &status, sizeof(int));
        if (status < 0)
            return 1;
        status = execvp(argv[0], argv);
        exit(status);
    }

    MPI_Allreduce(&status, &global_status, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
    if (global_status < 0){
        if (rank == 0){
            fprintf(stderr, "Warning: 1 or more nodes was unable to "
                    "complete the operation");
        }
    }

    MPI_Finalize();
    return 0;
}
